/* Copyright (c) 1993
 *      Juergen Weigert (jnweiger@immd4.informatik.uni-erlangen.de)
 *      Michael Schroeder (mlschroe@immd4.informatik.uni-erlangen.de)
 * Copyright (c) 1987 Oliver Laumann
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program (see the file COPYING); if not, write to the
 * Free Software Foundation, Inc.,
 * 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA
 *
 ****************************************************************
 */

#include "rcs.h"
RCS_ID("$Id: window.c,v 1.20 1994/05/31 12:33:24 mlschroe Exp $ FAU")

#include <sys/types.h>
#include <sys/stat.h>
#include <signal.h>
#include <fcntl.h>
#ifndef sun
#include <sys/ioctl.h>
#endif

#include "config.h"

/* for solaris 2.1, Unixware (SVR4.2) and possibly others: */
#ifdef SVR4
# include <sys/stropts.h>
#endif

#include "screen.h"
#include "extern.h"
#include "logfile.h"	/* logfopen() */

extern struct display *displays, *display;
extern struct win *windows, *fore, *wtab[], *console_window;
extern char *ShellArgs[];
extern char *ShellProg;
extern char screenterm[];
extern char *screenlogfile;
extern char HostName[];
extern int TtyMode;
extern int SilenceWait;
extern struct LayFuncs WinLf;
extern int real_uid, real_gid, eff_uid, eff_gid;
extern char Termcap[];
extern char **NewEnv;

#if defined(TIOCSWINSZ) || defined(TIOCGWINSZ)
extern struct winsize glwz;
#endif

#ifdef _IBMR2
extern int aixhack;
#endif

static int OpenDevice __P((char *, char *, int, struct tty_attr *, char **));
static int ForkWindow __P((char **, char *, char *, char *, struct win *));
static void execvpe __P((char *, char **, char **));


int VerboseCreate = 0;		/* XXX move this to user.h */

char DefaultShell[] = "/bin/sh";
static char DefaultPath[] = ":/usr/ucb:/bin:/usr/bin";

/* keep this in sync with the structure definition in window.h */
struct NewWindow nwin_undef   = 
{
  -1,		/* StartAt */
  (char *)0,	/* aka */
  (char **)0,	/* args */
  (char *)0,	/* dir */
  (char *)0,	/* term */
  -1,		/* aflag */
  -1,		/* flowflag */
  -1,		/* lflag */
  -1,		/* histheight */
  -1,		/* monitor */
  -1,		/* wlock */
  -1,		/* silence */
  -1,		/* wrap */
  -1,		/* logging */
  -1,		/* slowpaste */
  -1, 		/* c1 */
  -1, 		/* gr */
  -1, 		/* kanji */
  (char *)0,	/* hstatus */
  (char *)0	/* charset */
};

struct NewWindow nwin_default = 
{ 
  0, 		/* StartAt */
  0, 		/* aka */
  ShellArgs, 	/* args */
  0, 		/* dir */
  screenterm, 	/* term */
  0, 		/* aflag */
  1*FLOW_NOW,	/* flowflag */ 
  LOGINDEFAULT, /* lflag */
  DEFAULTHISTHEIGHT, 	/* histheight */
  MON_OFF, 	/* monitor */
  WLOCK_AUTO, 	/* wlock */
  -1, 		/* silence */
  1,		/* wrap */
  0,		/* logging */
  0,		/* slowpaste */
  1,		/* c1 */
  0,		/* gr */
  0,		/* kanji */
  (char *)0,	/* hstatus */
  (char *)0	/* charset */
};

struct NewWindow nwin_options;

void
nwin_compose(def, new, res)
struct NewWindow *def, *new, *res;
{
#define COMPOSE(x) res->x = new->x != nwin_undef.x ? new->x : def->x
  COMPOSE(StartAt);
  COMPOSE(aka);
  COMPOSE(args);
  COMPOSE(dir);
  COMPOSE(term);
  COMPOSE(aflag);
  COMPOSE(flowflag);
  COMPOSE(lflag);
  COMPOSE(histheight);
  COMPOSE(monitor);
  COMPOSE(wlock);
  COMPOSE(silence);
  COMPOSE(wrap);
  COMPOSE(Lflag);
  COMPOSE(slow);
  COMPOSE(c1);
  COMPOSE(gr);
  COMPOSE(kanji);
  COMPOSE(hstatus);
  COMPOSE(charset);
#undef COMPOSE
}
 
/*
 * DoStartLog constructs a path for the "want to be logfile" in buf and 
 * attempts logfopen.
 *
 * needs char buf[1024];
 * returns 0 on success.
 */
int 
DoStartLog(w, buf)
struct win *w;
char *buf;
{
  if (!w || !buf)
    return -1;

  strncpy(buf, MakeWinMsg(screenlogfile, w, '%'), 1023);
  buf[1023] = 0;

  debug2("DoStartLog: win %d, file %s\n", w->w_number, buf);

  if (w->w_log != NULL)
    logfclose(w->w_log);

  if ((w->w_log = logfopen(buf, islogfile(buf) ? NULL : secfopen(buf, "a"))) == NULL)
    return -2;
  return 0;
}

/* 
 * Umask & wlock are set for the user of the display, 
 * The display d (if specified) switches to that window.
 */
int
MakeWindow(newwin)
struct NewWindow *newwin;
{
  register struct win **pp, *p;
  register int n, i;
  int f = -1;
  struct NewWindow nwin;
  struct tty_attr tattr;
  char *TtyName, *TtyOpt = NULL;
#ifdef MULTIUSER
  extern struct user *users;
#endif

  debug1("NewWindow: StartAt %d\n", newwin->StartAt);
  debug1("NewWindow: aka     %s\n", newwin->aka?newwin->aka:"NULL");
  debug1("NewWindow: dir     %s\n", newwin->dir?newwin->dir:"NULL");
  debug1("NewWindow: term    %s\n", newwin->term?newwin->term:"NULL");

  nwin_compose(&nwin_default, newwin, &nwin);
  debug1("NWin: aka     %s\n", nwin.aka ? nwin.aka : "NULL");
  debug1("NWin: wlock   %d\n", nwin.wlock);
  debug1("NWin: Lflag   %d\n", nwin.Lflag);

  pp = wtab + nwin.StartAt;

  do
    {
      if (*pp == 0)
	break;
      if (++pp == wtab + MAXWIN)
	pp = wtab;
    }
  while (pp != wtab + nwin.StartAt);
  if (*pp)
    {
      Msg(0, "No more windows.");
      return -1;
    }

#if defined(USRLIMIT) && defined(UTMPOK)
  /*
   * Count current number of users, if logging windows in.
   */
  if (nwin.lflag && CountUsers() >= USRLIMIT)
    {
      Msg(0, "User limit reached.  Window will not be logged in.");
      nwin.lflag = 0;
    }
#endif
  n = pp - wtab;
  debug1("Makewin creating %d\n", n);

  if ((f = OpenDevice(nwin.args[0], nwin.args[1], nwin.lflag, &tattr, &TtyName)) < 0)
    return -1;
  if (tattr.flag == TTY_FLAG_PLAIN)
    TtyOpt = nwin.args[1];

  if ((p = (struct win *) malloc(sizeof(struct win))) == 0)
    {
      close(f);
      Msg(0, strnomem);
      return -1;
    }
  bzero((char *) p, (int) sizeof(struct win)); /* looks like a calloc above */

  /* save the command line so that zombies can be resurrected */
  for (i = 0; nwin.args[i] && i < MAXARGS - 1; i++)
    p->w_cmdargs[i] = SaveStr(nwin.args[i]);
  p->w_cmdargs[i] = 0;

  p->w_number = n;
#ifdef MULTIUSER
  /* 
   * This is dangerous: without a display we use creators umask 
   * This is intended to be usefull for detached startup.
   * But is still better than default bits with a NULL user.
   */
  if (NewWindowAcl(p, display ? D_user : users))
    {
      free((char *)p);
      close(f);
      Msg(0, strnomem);
      return -1;
    }
#endif
  p->w_winlay.l_next = 0;
  p->w_winlay.l_layfn = &WinLf;
  p->w_winlay.l_data = (char *)p;
  p->w_lay = &p->w_winlay;
  p->w_display = display;
  p->w_pdisplay = 0;

#ifdef MULTIUSER
  if (display && !AclCheckPermWin(D_user, ACL_WRITE, p))
#else
  if (display)
#endif
    p->w_wlockuser = D_user;
  p->w_wlock = nwin.wlock;
  p->w_ptyfd = f;
  p->w_aflag = nwin.aflag;
  p->w_flow = nwin.flowflag | ((nwin.flowflag & FLOW_AUTOFLAG) ? (FLOW_AUTO|FLOW_NOW) : FLOW_AUTO);
  if (!nwin.aka)
    nwin.aka = Filename(nwin.args[0]);
  strncpy(p->w_akabuf, nwin.aka, MAXSTR - 1);
  if ((nwin.aka = rindex(p->w_akabuf, '|')) != NULL)
    {
      p->w_autoaka = 0;
      *nwin.aka++ = 0;
      p->w_title = nwin.aka;
      p->w_akachange = nwin.aka + strlen(nwin.aka);
    }
  else
    p->w_title = p->w_akachange = p->w_akabuf;
  if (nwin.hstatus)
    p->w_hstatus = SaveStr(nwin.hstatus);
  p->w_monitor = nwin.monitor;
  /*
   * defsilence by Lloyd Zusman (zusman_lloyd@jpmorgan.com)
   */
  p->w_silence = nwin.silence;
  if (p->w_silence > 0)
    {
      if (p->w_tstamp.seconds != 0)
        {
	  p->w_tstamp.lastio = time(0);
	  p->w_tstamp.seconds = SilenceWait;
	}
      else
        {
	  p->w_tstamp.lastio = (time_t)0;
	  p->w_tstamp.seconds = 0;
	}
     }
  p->w_slowpaste = nwin.slow;

  if (TtyOpt && *TtyOpt)
    {
      i = strlen(p->w_tty) + 1;
      p->w_tty_opt = p->w_tty + i;
      strncpy(p->w_tty_opt, TtyOpt, MAXSTR - 1 - i);
    }
  p->w_norefresh = 0;
  strncpy(p->w_tty, *TtyName ? TtyName : p->w_title, MAXSTR - 1);

#ifndef COPY_PASTE
  nwin.histheight = 0;
#endif
  if (ChangeWindowSize(p, display ? D_defwidth : 80,
		       display ? D_defheight : 24, 
		       nwin.histheight))
    {
      FreeWindow(p);
      return -1;
    }
#ifdef KANJI
  p->w_kanji = nwin.kanji;
#endif
  ResetWindow(p);	/* sets w_wrap, w_c1, w_gr */
  if (nwin.charset)
    SetCharsets(p, nwin.charset);

  /* same verbosity as in RemakeWindow() ... */
  if (tattr.flag == TTY_FLAG_PLAIN)
    {
      if (VerboseCreate)
	{
	  struct display *d = display; /* WriteString zaps display */

	  WriteString(p, p->w_cmdargs[0], strlen(p->w_cmdargs[0]));
	  WriteString(p, ": ", 2);
	  WriteString(p, p->w_title, strlen(p->w_title));
	  WriteString(p, "\r\n", 2);
	  display = d;
	}
      p->w_t = tattr;
      p->w_pid = 0;
    }
  else
    {
      debug("forking...\n");
#ifdef PSEUDOS
      p->w_pwin = NULL;
#endif
      if (VerboseCreate)
	{
	  struct display *d = display; /* WriteString zaps display */

	  WriteString(p, ":screen", 7);	/* to show where we came from */
	  for (f = 0; p->w_cmdargs[f]; f++)
	    {
	      WriteString(p, " ", 1);
	      WriteString(p, p->w_cmdargs[f], strlen(p->w_cmdargs[f]));
	    }
	  WriteString(p, "\r\n", 2);
	  display = d;
	}
      p->w_pid = ForkWindow(nwin.args, nwin.dir, nwin.term, TtyName, p);
      if (p->w_pid < 0)
	{
	  FreeWindow(p);
	  return -1;
	}
    }
  /*
   * Place the newly created window at the head of the most-recently-used list.
   */
  if (display && D_fore)
    D_other = D_fore;
  *pp = p;
  p->w_next = windows;
  windows = p;
#ifdef UTMPOK
  p->w_slot = SLOT_LOGIN_NEVER;
# ifdef LOGOUTOK
  debug1("MakeWindow will %slog in.\n", nwin.lflag?"":"not ");
  if (nwin.lflag)
# else /* LOGOUTOK */
  debug1("MakeWindow will log in, LOGOUTOK undefined in config.h%s.\n",
  	 nwin.lflag?"":" (although lflag=0)");
# endif /* LOGOUTOK */
    {
      p->w_slot = SLOT_LOGIN_WHEN_REATTACH;	/* SLOT_LOGIN_NOT_YET */
      if (display)
        SetUtmp(p);
    }
# ifdef CAREFULUTMP
  CarefulUtmp();		/* If all 've been zombies, we've had no slot */
# endif
#endif /* UTMPOK */

  if (nwin.Lflag)
    {
      char buf[1024];

      DoStartLog(p, buf);
    }

  SetForeWindow(p);
  Activate(p->w_norefresh);
  return n;
}

/*
 * Resurrect a window from Zombie state. 
 * The command vector is therefore stored in the window structure.
 * Note: The terminaltype defaults to screenterm again, the current 
 * working directory is lost.
 */
int
RemakeWindow(p)
struct win *p;
{
  struct tty_attr tattr;
  char *TtyName;
  int lflag, f;

  lflag = p->w_slot != SLOT_LOGIN_NEVER;
  if ((f = OpenDevice(p->w_cmdargs[0], p->w_cmdargs[1], lflag, &tattr, &TtyName)) < 0)
    return -1;

  strncpy(p->w_tty, *TtyName ? TtyName : p->w_title, MAXSTR - 1);
  p->w_ptyfd = f;

  /* same verbosity as in MakeWindow() ... */
  if (tattr.flag == TTY_FLAG_PLAIN)
    {
      if (VerboseCreate)
	{
	  WriteString(p, p->w_cmdargs[0], strlen(p->w_cmdargs[0]));
	  WriteString(p, ": ", 2);
	  WriteString(p, p->w_title, strlen(p->w_title));
	  WriteString(p, "\r\n", 2);
	}
      p->w_t = tattr;
      p->w_pid = 0;
    }
  else 
    {
      if (VerboseCreate)
	{
	  WriteString(p, ":screen", 7);	/* to show where we came from */
	  for (f = 0; p->w_cmdargs[f]; f++)
	    {
	      WriteString(p, " ", 1);
	      WriteString(p, p->w_cmdargs[f], strlen(p->w_cmdargs[f]));
	    }
	  WriteString(p, "\r\n", 2);
	}
      p->w_pid = ForkWindow(p->w_cmdargs, (char *)0, nwin_default.term, TtyName, p);
      if (p->w_pid < 0)
	return -1;
    }

#ifdef UTMPOK
  p->w_slot = SLOT_LOGIN_NEVER;
  debug1("RemakeWindow will %slog in.\n", lflag ? "" : "not ");
# ifdef LOGOUTOK
  if (lflag)
# endif
    {
      p->w_slot = SLOT_LOGIN_WHEN_REATTACH;	/* SLOT_LOGIN_NOT_YET */
      if (display)
	SetUtmp(p);
    }
# ifdef CAREFULUTMP
  CarefulUtmp();		/* If all 've been zombies, we've had no slot */
# endif
#endif
  return p->w_number;
}

void
CloseDevice(wp)
struct win *wp;
{
  if (wp->w_ptyfd >= 0)
    {
      if (wp->w_t.flag == TTY_FLAG_FORK)
        {
	  /* pty 4 SALE */
	  (void) chmod(wp->w_tty, 0666);
	  (void) chown(wp->w_tty, 0, 0);
	}
      else
	{
	  /* carefully restore what was previously there */
#ifdef HAVE_FCHMOD
	  (void) fchmod(wp->w_ptyfd,
#else
	  if (!strncmp(wp->w_tty, "/dev", 4))
	  (void) chmod(wp->w_tty, 
#endif
	    (wp->w_t.flag == TTY_FLAG_FORK) ? 0666 : wp->w_t.mode & 0777);

#ifdef HAVE_FCHMOD
	  (void) fchown(wp->w_ptyfd,
#else
	  if (!strncmp(wp->w_tty, "/dev", 4))
	  (void) chown(wp->w_tty, 
#endif
	    (wp->w_t.flag == TTY_FLAG_FORK) ? 0 : wp->w_t.uid, 
	    (wp->w_t.flag == TTY_FLAG_FORK) ? 0 : wp->w_t.gid);
	}
       close(wp->w_ptyfd);
       wp->w_ptyfd = -1;
     }
}

void
FreeWindow(wp)
struct win *wp;
{
  struct display *d;
  int i;

  debug1("FreeWindow %d\n", wp ? wp->w_number: -1);
#ifdef PSEUDOS
  if (wp->w_pwin)
    FreePseudowin(wp);
#endif
#ifdef UTMPOK
  RemoveUtmp(wp);
#endif
  CloseDevice(wp);

  if (wp == console_window)
    console_window = 0;
  if (wp->w_log != NULL)
    logfclose(wp->w_log);
  ChangeWindowSize(wp, 0, 0, 0);

  if (wp->w_hstatus)
    free(wp->w_hstatus);
  for (i = 0; wp->w_cmdargs[i]; i++)
    free(wp->w_cmdargs[i]);
  for (d = displays; d; d = d->d_next)
    {
      if (d->d_other == wp)
        d->d_other = NULL;
      if (d->d_fore == wp)
        d->d_fore = NULL;
    }

#ifdef MULTIUSER
  for (i = 0; i < ACL_BITS_PER_WIN; i++)
    free((char *)wp->w_userbits[i]);
  free((char *)wp->w_lio_notify);
  free((char *)wp->w_mon_notify);
#endif /* MULTIUSER */
  free((char *)wp);
}

static int
OpenDevice(arg, tty_opt, lflag, tattr, namep)
char *arg;
char *tty_opt;
int lflag;
struct tty_attr *tattr;
char **namep;
{
  struct stat st;
  int f;

  if ((stat(arg, &st)) == 0 && S_ISCHR(st.st_mode))
    {
      if (access(arg, R_OK | W_OK) == -1)
	{
	  Msg(errno, "Cannot access line '%s' for R/W", arg); 
	  return -1;
	}
      debug("OpenDevice: OpenTTY\n");
      if ((f = OpenTTY(arg, tty_opt)) < 0)
	return -1;
      tattr->flag = TTY_FLAG_PLAIN;
      tattr->uid = st.st_uid;
      tattr->gid = st.st_gid;
      tattr->mode = st.st_mode;
      lflag = 0;
      *namep = arg;
    }
  else
    {
      tattr->flag = TTY_FLAG_FORK;    /* for now we hope it is a program */
      tattr->uid = tattr->gid = 0;
      tattr->mode = 0666;
      f = OpenPTY(namep);
      if (f == -1)
	{
	  Msg(0, "No more PTYs.");
	  return -1;
	}
#ifdef TIOCPKT
      {
	int flag = 1;

	if (ioctl(f, TIOCPKT, (char *)&flag))
	  {
	    Msg(errno, "TIOCPKT ioctl");
	    close(f);
	    return -1;
	  }
      }
#endif /* TIOCPKT */
    }
#ifdef TTY_DISABLE_FNBLOCK
  if (tattr->flag == TTY_FLAG_PLAIN)
    {
      debug1("fcntl(%d, F_SETFL, 0)", f);
      (void) fcntl(f, F_SETFL, 0);
    }
  else
#endif
    {
      debug1("fcntl(%d, F_SETFL, FNBLOCK)\n", f);
      (void) fcntl(f, F_SETFL, FNBLOCK);
    }
#ifdef PTYGROUP
  (void) chown(*namep, real_uid, PTYGROUP);
#else
  (void) chown(*namep, real_uid, real_gid);
#endif
#ifdef UTMPOK
  (void) chmod(*namep, lflag ? TtyMode : (TtyMode & ~022));
#else
  (void) chmod(*namep, TtyMode);
#endif
  return f;
}

/*
 * Fields w_width, w_height, aflag, number (and w_tty)
 * are read from struct win *win. No fields written.
 * If pwin is nonzero, filedescriptors are distributed 
 * between win->w_tty and open(ttyn)
 *
 */
static int 
ForkWindow(args, dir, term, ttyn, win)
char **args, *dir, *term, *ttyn;
struct win *win;
{
  int pid;
  char tebuf[25];
  char ebuf[10];
  char shellbuf[7 + MAXPATHLEN];
  char *proc;
#ifndef TIOCSWINSZ
  char libuf[20], cobuf[20];
#endif
  int newfd;
  int w = win->w_width;
  int h = win->w_height;
#ifdef PSEUDOS
  int i, pat, wfdused;
  struct pseudowin *pwin = win->w_pwin;
#endif

  proc = *args;
  if (proc == 0)
    {
      args = ShellArgs;
      proc = *args;
    }
  switch (pid = fork())
    {
    case -1:
      Msg(errno, "fork");
      break;
    case 0:
      signal(SIGHUP, SIG_DFL);
      signal(SIGINT, SIG_DFL);
      signal(SIGQUIT, SIG_DFL);
      signal(SIGTERM, SIG_DFL);
#ifdef BSDJOBS
      signal(SIGTTIN, SIG_DFL);
      signal(SIGTTOU, SIG_DFL);
#endif
#ifdef SIGPIPE
      signal(SIGPIPE, SIG_DFL);
#endif
#ifdef SIGXFSZ
      signal(SIGXFSZ, SIG_DFL);
#endif

      displays = 0;		/* beware of Panic() */
      if (setuid(real_uid) || setgid(real_gid))
	{
	  SendErrorMsg("Setuid/gid: %s", strerror(errno));
	  exit(1);
	}
      eff_uid = real_uid;
      eff_gid = real_gid;
      if (dir && *dir && chdir(dir) == -1)
	{
	  SendErrorMsg("Cannot chdir to %s: %s", dir, strerror(errno));
	  exit(1);
	}

      if (display)
	{
	  brktty(D_userfd);
	  freetty();
	}
      else
	brktty(-1);
#ifdef DEBUG
      if (dfp && dfp != stderr)
	fclose(dfp);
#endif
#ifdef _IBMR2
      close(0);
      dup(aixhack);
      close(aixhack);
#endif
      closeallfiles(win->w_ptyfd);
#ifdef _IBMR2
      aixhack = dup(0);
#endif
#ifdef DEBUG
      if (dfp)	/* do not produce child debug, when debug is "off" */
	{
	  char buf[256];
	  
	  sprintf(buf, "%s/screen.child", DEBUGDIR);
	  if ((dfp = fopen(buf, "a")) == 0)
	    dfp = stderr;
	  else
	    (void) chmod(buf, 0666);
	}
      debug1("=== ForkWindow: pid %d\n", (int)getpid());
#endif
      /* Close the three /dev/null descriptors */
      close(0);
      close(1);
      close(2);
      newfd = -1;
      /* 
       * distribute filedescriptors between the ttys
       */
#ifdef PSEUDOS
      pat = pwin ? pwin->fdpat : 
		   ((F_PFRONT<<(F_PSHIFT*2)) | (F_PFRONT<<F_PSHIFT) | F_PFRONT);
      wfdused = 0;
      for(i = 0; i < 3; i++)
	{
	  if (pat & F_PFRONT << F_PSHIFT * i)
	    {
	      if (newfd < 0)
		{
		  if ((newfd = open(ttyn, O_RDWR)) < 0)
		    {
		      SendErrorMsg("Cannot open %s: %s", ttyn, strerror(errno));
		      exit(1);
		    }
		}
	      else
		dup(newfd);
	    }
	  else
	    {
	      dup(win->w_ptyfd);
	      wfdused = 1;
	    }
	}
      if (wfdused)
	{
	    /* 
	     * the pseudo window process should not be surprised with a 
	     * nonblocking filedescriptor. Poor Backend!
	     */
	    debug1("Clearing NBLOCK on window-fd(%d)\n", win->w_ptyfd);
	    if (fcntl(win->w_ptyfd, F_SETFL, 0))
	      SendErrorMsg("Warning: ForkWindow clear NBLOCK fcntl failed, %d", errno);
	}
#else /* PSEUDOS */
      if ((newfd = open(ttyn, O_RDWR)) != 0)
	{
	  SendErrorMsg("Cannot open %s: %s", ttyn, strerror(errno));
	  exit(1);
	}
      dup(0);
      dup(0);
#endif /* PSEUDOS */
      close(win->w_ptyfd);
#ifdef _IBMR2
      close(aixhack);
#endif

      if (newfd >= 0)
	{
	  struct mode fakemode, *modep;
#ifdef I_PUSH	/* SINIX is as bad as sgi :-) */
# if defined(SVR4) && !defined(sgi)
	  if (ioctl(newfd, I_PUSH, "ptem"))
	    {
	      SendErrorMsg("Cannot I_PUSH ptem %s %s", ttyn, strerror(errno));
	      exit(1);
	    }
	  if (ioctl(newfd, I_PUSH, "ldterm"))
	    {
	      SendErrorMsg("Cannot I_PUSH ldterm %s %s", ttyn, strerror(errno));
	      exit(1);
	    }
	  if (ioctl(newfd, I_PUSH, "ttcompat"))
	    {
	      SendErrorMsg("Cannot I_PUSH ttcompat %s %s", ttyn, strerror(errno));
	      exit(1);
	    }
# endif
#endif
	  if (fgtty(newfd))
	    SendErrorMsg("fgtty: %s (%d)", strerror(errno), errno);
	  if (display)
	    {
	      debug("ForkWindow: using display tty mode for new child.\n");
	      modep = &D_OldMode;
	    }
	  else
	    {
	      debug("No display - creating tty setting\n");
	      modep = &fakemode;
	      InitTTY(modep, 0);
#ifdef DEBUG
	      DebugTTY(modep);
#endif
	    }
	  /* We only want echo if the users input goes to the pseudo
	   * and the pseudo's stdout is not send to the window.
	   */
#ifdef PSEUDOS
	  if (pwin && (!(pat & F_UWP) || (pat & F_PBACK << F_PSHIFT)))
	    {
	      debug1("clearing echo on pseudywin fd (pat %x)\n", pat);
# if defined(POSIX) || defined(TERMIO)
	      modep->tio.c_lflag &= ~ECHO;
	      modep->tio.c_iflag &= ~ICRNL;
# else
	      modep->m_ttyb.sg_flags &= ~ECHO;
# endif
	    }
#endif
	  SetTTY(newfd, modep);
#ifdef TIOCSWINSZ
	  glwz.ws_col = w;
	  glwz.ws_row = h;
	  (void) ioctl(newfd, TIOCSWINSZ, (char *)&glwz);
#endif
	  /* Always turn off nonblocking mode */
	  (void)fcntl(newfd, F_SETFL, 0);
	}
#ifndef TIOCSWINSZ
      sprintf(libuf, "LINES=%d", h);
      sprintf(cobuf, "COLUMNS=%d", w);
      NewEnv[5] = libuf;
      NewEnv[6] = cobuf;
#endif
#ifdef MAPKEYS
      NewEnv[2] = MakeTermcap(display == 0 || win->w_aflag);
#else
      if (win->w_aflag)
	NewEnv[2] = MakeTermcap(1);
      else
	NewEnv[2] = Termcap;
#endif
      strcpy(shellbuf, "SHELL=");
      strncpy(shellbuf + 6, ShellProg, MAXPATHLEN);
      shellbuf[MAXPATHLEN + 6] = 0;
      NewEnv[4] = shellbuf;
      debug1("ForkWindow: NewEnv[4] = '%s'\n", shellbuf);
      if (term && *term && strcmp(screenterm, term) &&
	  (strlen(term) < 20))
	{
	  char *s1, *s2, tl;

	  sprintf(tebuf, "TERM=%s", term);
	  debug2("Makewindow %d with %s\n", win->w_number, tebuf);
	  tl = strlen(term);
	  NewEnv[1] = tebuf;
	  if ((s1 = index(NewEnv[2], '|')))
	    {
	      if ((s2 = index(++s1, '|')))
		{
		  if (strlen(NewEnv[2]) - (s2 - s1) + tl < 1024)
		    {
		      bcopy(s2, s1 + tl, strlen(s2) + 1);
		      bcopy(term, s1, tl);
		    }
		}
	    }
	}
      sprintf(ebuf, "WINDOW=%d", win->w_number);
      NewEnv[3] = ebuf;

      if (*proc == '-')
	proc++;
      if (!*proc)		/* andreas@planix.com: if args[0] = "" */
        proc = DefaultShell;
      debug1("calling execvpe %s\n", proc);
      execvpe(proc, args, NewEnv);
      debug1("exec error: %d\n", errno);
      SendErrorMsg("Cannot exec \"%s\": %s", proc, strerror(errno));
      exit(1);
    default:
      break;
    }
#ifdef _IBMR2
  close(aixhack);
  aixhack = -1;
#endif
  return pid;
}

static void
execvpe(prog, args, env)
char *prog, **args, **env;
{
  register char *path = NULL, *p;
  char buf[1024];
  char *shargs[MAXARGS + 1];
  register int i, eaccess = 0;

  if (rindex(prog, '/'))
    path = "";
  if (!path && !(path = getenv("PATH")))
    path = DefaultPath;
  do
    {
      p = buf;
      while (*path && *path != ':')
	*p++ = *path++;
      if (p > buf)
	*p++ = '/';
      strcpy(p, prog);
      execve(buf, args, env);
      switch (errno)
	{
	case ENOEXEC:
	  shargs[0] = DefaultShell;
	  shargs[1] = buf;
	  for (i = 1; (shargs[i + 1] = args[i]) != NULL; ++i)
	    ;
	  execve(DefaultShell, shargs, env);
	  return;
	case EACCES:
	  eaccess = 1;
	  break;
	case ENOMEM:
	case E2BIG:
	case ETXTBSY:
	  return;
	}
    } while (*path++);
  if (eaccess)
    errno = EACCES;
}

#ifdef PSEUDOS

int
winexec(av)
char **av;
{
  char **pp;
  char *p, *s, *t;
  int i, r = 0, l = 0;
  struct win *w;
  extern struct display *display;
  extern struct win *windows;
  struct pseudowin *pwin;
  struct tty_attr tattr;
  
  if ((w = display ? fore : windows) == NULL)
    return -1;
  if (!*av || w->w_pwin)
    {
      Msg(0, "Filter running: %s", w->w_pwin ? w->w_pwin->p_cmd : "(none)");
      return -1;
    }
  if (w->w_ptyfd < 0)
    {
      Msg(0, "You feel dead inside.");
      return -1;
    }
  if (!(pwin = (struct pseudowin *)malloc(sizeof(struct pseudowin))))
    {
      Msg(0, strnomem);
      return -1;
    }
  bzero((char *)pwin, (int)sizeof(*pwin));

  /* allow ^a:!!./ttytest as a short form for ^a:exec !.. ./ttytest */
  for (s = *av; *s == ' '; s++)
    ;
  for (p = s; *p == ':' || *p == '.' || *p == '!'; p++)
    ;
  if (*p != '|')
    while (*p && p > s && p[-1] == '.')
      p--;
  if (*p == '|')
    {
      l = F_UWP;
      p++;
    }
  if (*p) 
    av[0] = p;
  else
    av++;

  t = pwin->p_cmd;
  for (i = 0; i < 3; i++)
    {
      *t = (s < p) ? *s++ : '.';
      switch (*t++)
	{
	case '.':
	case '|':
	  l |= F_PFRONT << (i * F_PSHIFT);
	  break;
	case '!':
	  l |= F_PBACK << (i * F_PSHIFT);
	  break;
	case ':':
	  l |= F_PBOTH << (i * F_PSHIFT);
	  break;
	}
    }
  
  if (l & F_UWP)
    {
      *t++ = '|';
      if ((l & F_PMASK) == F_PFRONT)
	{
	  *pwin->p_cmd = '!';
	  l ^= F_PFRONT | F_PBACK;
	}
    }
  if (!(l & F_PBACK))
    l |= F_UWP;
  *t++ = ' ';
  pwin->fdpat = l;
  debug1("winexec: '%#x'\n", pwin->fdpat);
  
  l = MAXSTR - 4;
  for (pp = av; *pp; pp++)
    {
      p = *pp;
      while (*p && l-- > 0)
        *t++ = *p++;
      if (l <= 0)
	break;
      *t++ = ' ';
    }
  *--t = '\0';
  debug1("%s\n", pwin->p_cmd);
  
  if ((pwin->p_ptyfd = OpenDevice(av[0], NULL, 0, &tattr, &t)) < 0)
    {
      free((char *)pwin);
      return -1;
    }
  strncpy(pwin->p_tty, t, MAXSTR - 1);
  w->w_pwin = pwin;
  if (tattr.flag == TTY_FLAG_PLAIN)
    {
      FreePseudowin(w);
      Msg(0, "Cannot handle a TTY as a pseudo win.");
      return -1;
    }
#ifdef TIOCPKT
  {
    int flag = 0;

    if (ioctl(pwin->p_ptyfd, TIOCPKT, (char *)&flag))
      {
	Msg(errno, "TIOCPKT ioctl");
	FreePseudowin(w);
	return -1;
      }
  }
#endif /* TIOCPKT */
  pwin->p_pid = ForkWindow(av, NULL, NULL, t, w);
  if ((r = pwin->p_pid) < 0)
    FreePseudowin(w);
  return r;
}

void
FreePseudowin(w)
struct win *w;
{
  struct pseudowin *pwin = w->w_pwin;

  ASSERT(pwin);
  if (fcntl(w->w_ptyfd, F_SETFL, FNBLOCK))
    Msg(errno, "Warning: FreePseudowin: NBLOCK fcntl failed");
  /* should be able to use CloseDevice() here */
  (void) chmod(pwin->p_tty, 0666);
  (void) chown(pwin->p_tty, 0, 0);
  if (pwin->p_ptyfd >= 0)
    close(pwin->p_ptyfd);
  free((char *)pwin);
  w->w_pwin = NULL;
}

#endif /* PSEUDOS */


#ifdef MULTI
/*
 * Huch, what is that? Looks like a part of display.c
 */
/*
 *  Clone routines. To be removed...
 */

static int CloneTermcap __P((struct display *));
extern char **environ;
extern int iflag;

int
execclone(av)
char **av;
{
  int f, sf;
  char specialbuf[6];
  struct display *old = display;
  char **avp, *namep;

  sf = OpenPTY(&namep);
  if (sf == -1)
    {
      Msg(0, "No more PTYs.");
      return -1;
    }
#ifdef _IBMR2
  close(aixhack);
  aixhack = -1;
#endif
  f = open(namep, O_RDWR);
  if (f == -1)
    {
      close(sf);
      Msg(errno, "Cannot open slave");
      return -1;
    }
  brktty(f);
  signal(SIGHUP, SIG_IGN);	/* No hangups, please */
  if (MakeDisplay(D_username, namep, D_termname, f, -1, &D_OldMode) == 0)
    {
      display = old;
      Msg(0, "Could not make display.");
      close(f);
      close(sf);
      return -1;
    }
  if (CloneTermcap(old))
    {
      FreeDisplay();
      display = old;
      close(sf);
      return -1;
    }

  SetMode(&D_OldMode, &D_NewMode, D_flow, iflag);
  SetTTY(f, &D_NewMode);

  switch (fork())
    {
    case -1:
      FreeDisplay();
      display = old;
      Msg(errno, "fork");
      close(sf);
      return -1;
    case 0:
      D_usertty[0] = 0;		/* for SendErrorMsg */
      displays = 0;		/* beware of Panic() */
      if (setuid(real_uid) || setgid(real_gid))
	{
	  SendErrorMsg("Setuid/gid: %s", strerror(errno));
	  exit(1);
	}
      eff_uid = real_uid;
      eff_gid = real_gid;
      closeallfiles(sf);
      close(1);
      dup(sf);
      close(sf);
#ifdef DEBUG
	{
	  char buf[256];

	  sprintf(buf, "%s/screen.child", DEBUGDIR);
	  if ((dfp = fopen(buf, "a")) == 0)
	    dfp = stderr;
	  else
	    (void) chmod(buf, 0666);
	}
      debug1("=== Clone: pid %d\n", (int)getpid());
#endif
      for (avp = av; *avp; avp++)
        {
          if (strcmp(*avp, "%p") == 0)
            *avp = namep;
          if (strcmp(*avp, "%X") == 0)
            *avp = specialbuf;
        }
      sprintf(specialbuf, "-SXX1");
      namep += strlen(namep);
      specialbuf[3] = *--namep;
      specialbuf[2] = *--namep;
#ifdef DEBUG
      debug("Calling:");
      for (avp = av; *avp; avp++)
        debug1(" %s", *avp);
      debug("\n");
#endif
      execvpe(*av, av, environ);
      SendErrorMsg("Cannot exec \"%s\": %s", *av, strerror(errno));
      exit(1);
    default:
      break;
    }
  close(sf);
  InitTerm(0);
  Activate(0);
  if (D_fore == 0)
    ShowWindows();
  return 0;
}

extern struct term term[];      /* terminal capabilities */

static int
CloneTermcap(old)
struct display *old;
{
  char *tp;
  int i, l;

  l = 0;
  for (i = 0; i < T_N; i++)
    if (term[i].type == T_STR && old->d_tcs[i].str)
      l += strlen(old->d_tcs[i].str) + 1;
  if ((D_tentry = (char *)malloc(l)) == 0)
    {
      Msg(0, strnomem);
      return -1;
    }

  tp = D_tentry;
  for (i = 0; i < T_N; i++)
    {
      switch(term[i].type)
        {
        case T_FLG:
          D_tcs[i].flg = old->d_tcs[i].flg;
          break;
        case T_NUM:
          D_tcs[i].num = old->d_tcs[i].num;
          break;
        case T_STR:
          D_tcs[i].str = old->d_tcs[i].str;
          if (D_tcs[i].str)
            {
              strcpy(tp, D_tcs[i].str);
              D_tcs[i].str = tp;
              tp += strlen(tp) + 1;
            }
	  break;
        default:
          Panic(0, "Illegal tc type in entry #%d", i);
        }
    }
  CheckScreenSize(0);
  for (i = 0; i < NATTR; i++)
    D_attrtab[i] = old->d_attrtab[i];
  for (i = 0; i < 256; i++)
    D_c0_tab[i] = old->d_c0_tab[i];
  D_UPcost = old->d_UPcost;
  D_DOcost = old->d_DOcost;
  D_NLcost = old->d_NLcost;
  D_LEcost = old->d_LEcost;
  D_NDcost = old->d_NDcost;
  D_CRcost = old->d_CRcost;
  D_IMcost = old->d_IMcost;
  D_EIcost = old->d_EIcost;
#ifdef AUTO_NUKE
  D_auto_nuke = old->d_auto_nuke;
#endif
  if (D_CXC)
    CreateTransTable(D_CXC);
  D_tcinited = 1;
  return 0;
}
#endif /* MULTI */


/* 
 * returns 0, if the lock really has been released 
 */
int
ReleaseAutoWritelock(dis, w)
struct display *dis;
struct win *w;
{
  debug2("ReleaseAutoWritelock: user %s, window %d\n", 
         dis->d_user->u_name, w->w_number);

  /* release auto writelock when user has no other display here */
  if (w->w_wlock == WLOCK_AUTO && w->w_wlockuser == dis->d_user)
    {
      struct display *d;

      for (d = displays; d; d = d->d_next)
	if (( d != dis) && (d->d_fore == w) && (d->d_user == dis->d_user))
	  break;
      debug3("%s %s autolock on win %d\n", 
	     dis->d_user->u_name, d ? "keeps" : "releases", w->w_number);
      if (!d)
        {
	  w->w_wlockuser = NULL;
          return 0;
	}
    }
  return 1;
}

/* 
 * returns 0, if the lock really could be obtained 
 */
int
ObtainAutoWritelock(d, w)
struct display *d;
struct win *w;
{
  if ((w->w_wlock == WLOCK_AUTO) &&
#ifdef MULTIUSER
       !AclCheckPermWin(d->d_user, ACL_WRITE, w) &&
#endif
       !w->w_wlockuser)
    {
      debug2("%s obtained auto writelock for exported window %d\n",
             d->d_user->u_name, w->w_number);
      w->w_wlockuser = d->d_user;
      return 0;
    }
  return 1;
}
